# coding: utf-8
# pylint: disable=unused-argument
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import Iterable
from importlib import import_module
from itertools import groupby
from os.path import dirname
from os.path import join
import codecs
import contextlib

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db import connections
from django.db import router
from django.db.backends.signals import connection_created
from django.db.models.signals import post_migrate
from django.db.models.signals import pre_migrate
from django.db.transaction import atomic
from django.db.utils import ProgrammingError
from six import PY2
from six import iteritems
from six import itervalues
from six import text_type

from .constants import CORE_SCHEMA
from .constants import DATA_SCHEMA


class _Lock(object):

    u"""Блокировка с помощью функции PostgreSQL ``pg_advisory_lock``."""

    def __init__(self, connection):
        self._connection = connection

    def __enter__(self):
        with contextlib.closing(self._connection.cursor()) as cursor:
            cursor.callproc('pg_advisory_lock', (hash('m3-d15n'),))

    def __exit__(self, exc_type, exc_val, exc_tb):
        with contextlib.closing(self._connection.cursor()) as cursor:
            cursor.callproc('pg_advisory_unlock', (hash('m3-d15n'),))


class _SuspendConfigTriggers(object):

    def __init__(self, connection):
        self._connection = connection

    _sql = """
        alter table {core_schema}.columns_config
        {{action}} trigger {core_schema}_{{name}}
    """.format(core_schema=CORE_SCHEMA)

    def __enter__(self):
        with contextlib.closing(self._connection.cursor()) as cursor:
            cursor.execute(self._sql.format(
                action='disable',
                name='columns_config_changed',
            ))

    def __exit__(self, exc_type, exc_val, exc_tb):
        with contextlib.closing(self._connection.cursor()) as cursor:
            cursor.execute(self._sql.format(
                action='enable',
                name='columns_config_changed',
            ))


class ConnectionMixin(object):

    """Класс-примесь для работы с подключением к БД."""

    def __init__(self, database_alias):
        """Инициализация класса.

        :param basestring database_alias: Алиас деперсонализируемой базы
            данных.
        """
        self.alias = database_alias

    @property
    def _connection(self):
        """Подключение к базе данных.

        :rtype: django.db.backends.postgresql.base.DatabaseWrapper
        """
        result = connections[self.alias]
        return result

    @property
    def _cursor(self):
        """Курсор для выполнения запросов к БД.

        :rtype: django.db.backends.utils.CursorWrapper
        """
        result = self._connection.cursor()
        return result


class DatabaseParams(ConnectionMixin):

    """Класс для работы с параметрами деперсонализации."""

    defaults = dict(
        visible_chars_begin=2,
        visible_chars_end=0,
        result_length=10,
        fill_char='*',
    )

    def __init__(self, database_alias, defaults=None):
        """Инициализация класса.

        :param basestring database_alias: Алиас деперсонализируемой базы
            данных.
        """
        super(DatabaseParams, self).__init__(database_alias)

        if defaults:
            self.defaults = defaults
        else:
            self.defaults = DatabaseParams.defaults.copy()

        connection_created.connect(
            self._on_connection_created,
            dispatch_uid='m3_d15n.DatabaseParams.connection_created'
        )

    def _on_connection_created(self, connection, **kwargs):
        """При подключении устанавливает параметры подсистемы в БД.

        Т.к. параметры, установленные с помощью функции ``set_config``
        действуюют только в рамках текущего сеанса, необходимо при каждом
        подключении заново устанавливать значения всех параметров.
        """
        if connection.alias == self.alias:
            for param_name, param_value in iteritems(self.defaults):
                setattr(self, param_name, param_value)

    def _get_config(self, name, default_value=None):
        """Вызывает функцию ``current_setting`` в БД.

        :param base_string name: имя параметра.
        :param basestring default: значение параметра по умолчанию.
        """
        with self._cursor as cursor:
            try:
                cursor.callproc(CORE_SCHEMA + '.get_param', (name,))
            except ProgrammingError:
                result = default_value
            else:
                result = cursor.fetchone()[0]

        return result

    def _set_config(self, name, value):
        """Вызывает функцию ``set_config`` в БД.

        :param basestring name: имя параметра.
        :param basestring value: значение параметра.
        """
        with self._cursor as cursor:
            cursor.callproc('set_config', (name, text_type(value), False))

    @property
    def visible_chars_begin(self):
        """Количество видимых символов в начале строки.

        :rtype: int
        """
        result = max(0, int(
            self._get_config(
                'd15n_core.dn_visible_chars_begin',
                self.defaults['visible_chars_begin']
            )
        ))
        return result

    @visible_chars_begin.setter
    def visible_chars_begin(self, value):
        assert text_type(value).isdigit(), value

        self._set_config('d15n_core.dn_visible_chars_begin', text_type(value))

    @property
    def visible_chars_end(self):
        """Количество видимых символов в конце строки.

        :rtype: int
        """
        result = max(0, int(
            self._get_config(
                'd15n_core.dn_visible_chars_end',
                self.defaults['visible_chars_end']
            )
        ))
        return result

    @visible_chars_end.setter
    def visible_chars_end(self, value):
        assert text_type(value).isdigit(), value

        self._set_config('d15n_core.dn_visible_chars_end', text_type(value))

    @property
    def result_length(self):
        """Длина деперсонализированной строки.

        Если равна 0, то длина деперсонализированной строки будет равна
        длине исходной строки.

        :rtype: int
        """
        result = max(0, int(
            self._get_config(
                'd15n_core.dn_result_length',
                self.defaults['result_length']
            )
        ))
        return result

    @result_length.setter
    def result_length(self, value):
        assert text_type(value).isdigit(), value

        self._set_config('d15n_core.dn_result_length', text_type(value))

    @property
    def fill_char(self):
        """Символ-заполнитель.

        :rtype: unicode
        """
        result = self._get_config(
            'd15n_core.dn_fill_char',
            self.defaults['fill_char']
        )
        return result

    @fill_char.setter
    def fill_char(self, value):
        self._set_config('d15n_core.dn_fill_char', text_type(value))


class D15nMode(object):

    """Менеджер контекста, переводящий БД в режим деперсонализации.

    При включении выполняются следующие действия:

        - добавляет в параметр ``search_path`` текущей сессии PostgreSQL
          схему, в которой размещаются деперсонализирующие представления для
          таблиц Системы.

    .. note::

       Возможно использование вложенных менеджеров контекста.
    """

    def __init__(self, manager):  # noqa: D107
        self._manager = manager
        self._top_level = True

    def __enter__(self):  # noqa: D105
        self._top_level = not self._manager.is_in_d15n_mode()
        if self._top_level:
            self._manager.switch_on()

        return self

    def __exit__(self, exc_type, exc_value, traceback):  # noqa: D105
        if exc_type is None and self._top_level:
            self._manager.switch_off()


class D15nManager(ConnectionMixin):

    """Менеджер подсистемы деперсонализации.

    Предоставляет средства для установки и удаления из БД подсистемы
    деперсонализации, а также запроса сведений о её состоянии.
    """

    def __init__(self, database_alias, sync_app_params=False):  # noqa: D107
        super(D15nManager, self).__init__(database_alias)

        if self._connection.vendor != 'postgresql':
            raise ImproperlyConfigured(
                'Деперсонализация поддерживается только для СУБД PostgreSQL'
            )

        self.params = DatabaseParams(self.alias)

        self._config_synchronized = False
        self._sync_app_params = bool(sync_app_params)

        if self._sync_app_params:
            connection_created.connect(self._on_connection_created)

    @property
    def sync_app_params(self):
        """Признак автоматической синхронизации параметров Системы и БД.

        Значение ``True`` указывает на то, что при первом подключении к БД
        менеджер выполнит синхронизацию параметров деперсонализации,
        определенных в Системе, с параметрами, указанными в БД.

        :rtype: bool
        """
        return self._sync_app_params

    def _on_connection_created(self, connection, **kwargs):
        """Синхронизирует системные параметры деперсонализации с БД."""
        if connection.alias == self._connection.alias:
            self.sync_config()
            connection_created.disconnect(self._on_connection_created)

    def _execute_sql_file(self, file_name, **params):
        file_path = join(dirname(__file__), file_name)

        with codecs.open(file_path, 'r', 'utf-8') as sql_file:
            sql = sql_file.read()

        sql = sql.replace('{}', '{{}}')
        sql = sql.format(**params)

        with self._cursor as cursor:
            cursor.execute(sql)

    @atomic
    def install(self):
        """Устанавливает в БД необходимые для работы подсистемы объекты.

        Описание объектов (функций, таблиц, представлений, триггеров и т.п.)
        находится в файле ``sql/main.sql``.
        """
        self._execute_sql_file(
            'sql/main.sql',
            core_schema=CORE_SCHEMA,
            data_schema=DATA_SCHEMA,
        )

    def uninstall(self):
        """Удаляет из БД объекты подсистемы деперсонализации.

        Для этого удаляются схемы с именами, указанными в
        :obj:`~m3_d15n.constants.CORE_SCHEMA` и
        :obj:`~m3_d15n.constants.DATA_SCHEMA`.
        """
        with self._cursor as cursor:
            for schema in (DATA_SCHEMA, CORE_SCHEMA):
                cursor.execute(
                    'drop schema if exists {} cascade'.format(schema)
                )

    def _get_model_by_table_name(self, table_name):
        """Возвращает модель по имени таблицы в БД."""
        for model in apps.get_models(include_auto_created=True,
                                     include_swapped=True):
            # pylint: disable=protected-access
            if (
                model._meta.db_table == table_name and
                router.db_for_write(model) == self.alias
            ):
                return model

    def _get_search_path(self):
        """Возвращает список схем БД, в которых осуществляется поиск объектов.

        :rtype: list
        """
        with self._cursor as cursor:
            cursor.callproc(
                'current_setting',
                (
                    'search_path',
                ),
            )
            return cursor.fetchone()[0].split(', ')

    def _set_search_path(self, search_path):
        """Настраивает список схем БД, в которых осуществляется поиск объектов.

        :param list search_path: список схем БД.
        """
        with self._cursor as cursor:
            cursor.callproc(
                'set_config',
                (
                    'search_path',
                    ', '.join(search_path),
                    False
                ),
            )

    def is_in_d15n_mode(self):
        """Возвращает True, если включен режим деперсонализации.

        :rtype: bool
        """
        result = DATA_SCHEMA in self._get_search_path()
        return result

    def switch_on(self):
        """Включает режим деперсонализации в БД."""
        search_path = self._get_search_path()
        if DATA_SCHEMA not in search_path:
            search_path.insert(0, DATA_SCHEMA)
            self._set_search_path(search_path)

    def switch_off(self):
        """Выключает режим деперсонализации в БД."""
        search_path = self._get_search_path()
        if DATA_SCHEMA in search_path:
            search_path.remove(DATA_SCHEMA)
            self._set_search_path(search_path)

    @property
    def d15n_mode(self):
        """Менеджер контекста для перевода БД в режим деперсонализации.

        :rtype: m3_d15n.utils.D15nMode
        """
        result = D15nMode(self)
        return result

    def get_config(self):
        """Возвращает текущую конфигурацию подсистемы деперсонализации.

        :returns: Словарь вида

            .. code-block:: python

               {
                   model_class: (model_field1, model_field2, ...),
                   ...
               }
        :rtype: dict
        """
        result = {}

        with self._cursor as cursor:
            cursor.execute("""
                select
                    table_name,
                    column_name,
                    function_params
                from
                    {core_schema}.columns_config_as_names
                order by
                    table_name
            """.format(
                core_schema=CORE_SCHEMA,
            ))

            for table, fields in groupby(cursor, lambda row: row[0]):
                model = self._get_model_by_table_name(table)
                result[model] = tuple(
                    (field, field_params) for _, field, field_params in fields
                )

        return result

    def add_field(self, field, function_name, function_params=''):
        """Добавляет поле в параметры деперсонализации.

        :param field: поле модели.
        :type field: django.db.models.fields.Field

        :param basestring function_name: имя деперсонализирующей функции в БД,
            например ``'d15_core.depersonalize_name'``.

        :param basestring function_params: параметры деперсонализирующей
            функции. Строка вставляется в вызов функции "как есть", поэтому
            можно использовать любые выражения PostgreSQL.
        """
        with self._cursor as cursor:
            cursor.execute("""
                insert into {core_schema}.columns_config (
                    table_oid,
                    column_name,
                    function_oid,
                    function_params
                ) values (
                    'public.{table_name}'::regclass::oid,
                    '{column_name}',
                    to_regproc('{function_name}'),
                    %s
                )
            """.format(
                # pylint: disable=protected-access
                core_schema=CORE_SCHEMA,
                table_name=field.model._meta.db_table,
                column_name=field.column,
                function_name=function_name,
            ), [function_params])

    def delete_field(self, field):
        """Удаляет поле модели из параметров деперсонализации.

        :param field: поле модели
        :type field: django.db.models.fields.Field
        """
        # pylint: disable=protected-access
        with self._cursor as cursor:
            cursor.callproc(
                CORE_SCHEMA + '.get_table_schema',
                (
                    field.model._meta.db_table,
                )
            )
            table_schema = cursor.fetchone()[0]
            cursor.execute("""
                delete from {core_schema}.columns_config
                where
                    table_oid = '{table_schema}.{table_name}'::regclass and
                    column_name = '{column_name}'
            """.format(
                core_schema=CORE_SCHEMA,
                table_schema=table_schema,
                table_name=field.model._meta.db_table,
                column_name=field.column,
            ))

    def _collect_system_config(self):
        """Возвращает параметры деперсонализации, определенные в Системе.

        :param connection: Подключение к БД. Определяет, какие модели нужно
            учитывать при сборе параметров.

        :raises ValueError: если для одного и того же поля в разных приложениях
            заданы разные параметры деперсонализации.

        :returns: кортежи следующего вида:

            .. code-block:: python

               (app_label, model_name, field_name, func_name, func_params)
        :rtype: generator of tuple
        """
        for app in apps.get_app_configs():
            try:
                d15n_module = import_module(app.module.__name__ + '.d15n')
            except ImportError as error:
                if text_type(error).startswith('No module named '):
                    continue
                raise

            if not hasattr(d15n_module, 'field_params'):
                continue

            for model, fields_params in iteritems(d15n_module.field_params):
                assert isinstance(model, Iterable), model
                assert len(model) == 2, model
                model = apps.get_model(*model)
                if router.db_for_write(model) != self.alias:
                    continue

                for field_params in fields_params:
                    assert isinstance(field_params, Iterable), field_params
                    assert len(field_params) in (2, 3), field_params

                    # pylint: disable=protected-access
                    params = (
                        model._meta.app_label,
                        model._meta.object_name,
                        model._meta.get_field(field_params[0]).name,
                    )
                    if len(field_params) == 2:
                        params += (field_params[1], '')
                    elif len(field_params) == 3:
                        params += (field_params[1:])

                    yield params

    def _get_current_config(self):
        """Возвращает параметры деперсонализации, определенные в БД.

        :returns: кортежи следующего вида:

            .. code-block:: python

               (app_label, model_name, field_name, func_name, func_params)
        :rtype: generator
        """
        with self._connection.cursor() as cursor:
            cursor.execute("""
                select
                    table_name,
                    column_name,
                    function_schema,
                    function_name,
                    function_params
                from
                    d15n_core.columns_config_as_names
            """)
            for (
                table_name, column_name,
                function_schema, function_name, function_params
            ) in cursor:
                model = self._get_model_by_table_name(table_name)
                if model:
                    # pylint: disable=protected-access
                    yield (
                        model._meta.app_label,
                        model._meta.object_name,
                        model._meta.get_field(column_name).name,
                        function_schema + '.' + function_name,
                        function_params or ''
                    )

    def _add_fields_to_db(self, params):
        """Добавляет новые поля в конфигурацию БД."""
        # pylint: disable=protected-access
        for app, model, field, func_name, func_params in params:
            model = apps.get_model(app, model)
            field = model._meta.get_field(field)
            assert field.concrete, (app, model, field)

            self.add_field(field, func_name, func_params)

    def _delete_fields_from_db(self, params):
        """Удаляет поля из конфигурации БД."""
        # pylint: disable=protected-access
        for app, model, field, _, _ in params:
            model = apps.get_model(app, model)
            field = model._meta.get_field(field)
            assert field.concrete, (app, model, field)

            self.delete_field(field)

    @atomic
    def sync_config(self):
        """Выполняет синхронизацию параметров деперсонализации полей.

        На первом этапе синхронизации выполняется сбор параметров
        деперсонализации во всех подключенных к Системе приложениях. Для этого
        в каждом приложении осуществляется поиск модулей ``d15n.py``, в которых
        должен быть определен параметр ``FIALD_PARAMS``, содержащий словарь
        следующего вида:

        .. code-block:: python

           field_params = {
               ('person', 'Person'): (
                   ('surname', FUNC_DEPERSONALIZE_NAME),
                   ('firstname', FUNC_DEPERSONALIZE_NAME),
                   ('patronymic', FUNC_DEPERSONALIZE_NAME),
                   ('snils', FUNC_DEPERSONALIZE_NON_PUNCTUATION, "'x'"),
                   ('photo', FUNC_DEPERSONALIZE_TO_EMPTY_STRING),
                   ('photo_height', FUNC_DEPERSONALIZE_TO_NULL),
                   ('photo_width', FUNC_DEPERSONALIZE_TO_NULL),
                   ('act_address', FUNC_DEPERSONALIZE_TO_NULL),
                   ('reg_address', FUNC_DEPERSONALIZE_TO_NULL),
                   ('tmp_address', FUNC_DEPERSONALIZE_TO_NULL),
               ),
           }

        Затем осуществляется сверка параметров, определенных в Системе, с
        параметрами, указанными в БД (таблица ``columns_config``). При
        нахождении различий осуществляются соответствующие изменения в БД.
        """
        if self._config_synchronized:
            return

        with _Lock(self._connection):
            system_params = set(self._collect_system_config())
            db_params = set(self._get_current_config())

            if system_params != db_params:
                with _SuspendConfigTriggers(self._connection):
                    # Удаление столбцов, исключенных из конфигурации.
                    self._delete_fields_from_db(db_params - system_params)

                    # Добавление новых столбцов.
                    self._add_fields_to_db(system_params - db_params)

                with self._connection.cursor() as cursor:
                    cursor.callproc(CORE_SCHEMA + '.update_views_set')

        self._config_synchronized = True


class D15nManagerRegistry(object):

    """Реестр менеджеров деперсонализации."""

    def __init__(self):  # noqa: D107
        self._managers = {}
        self._skip_handler = None

    def skip_request(self, environ):
        """Проверяет, необходимо ли пропустить деперсонализацию запроса.

        Если задан обработчик запросов, проверка делегируется ему.

        :param environ: Словарь с параметрами запроса.
        """
        if self._skip_handler:
            return self._skip_handler(environ)

    def set_request_handler(self, handler):
        """Задает обработчик запросов.

        Обработчик должен быть callable-обьектом и принимать параметры
        запроса в виде словаря (параметр environ сигнала
        django.core.signals.request_started). В случае, если запрос должен быть
        деперсонализирован то обработчик должен возвращать None или False.

        :param handler: callable обьект
        """
        if handler and callable(handler):
            self._skip_handler = handler

    def register(self, manager):
        """Регистрирует менеджер деперсонализации в реестре.

        :param manager: m3_d15n.utils.D15nManager
        """
        assert isinstance(manager, D15nManager), type(manager)

        self._managers[manager.alias] = manager

    def get(self, database_alias):
        """Возвращает менеджер деперсонализации для указанной БД.

        :param basestring database_alias: Алиас базы данных, для которой
            запрашивается менеджер.

        :rtype: m3_d15n.utils.D15nManager
        """
        return self._managers[database_alias]

    def __contains__(self, database_alias):
        """Возвращает True, если для указанной БД зарегистрирован менеджер.

        :param basestring database_alias: Алиас базы данных.

        :rtype: bool
        """
        return database_alias in self._managers

    def is_in_d15n_mode(self):
        """Возвращает True, если все базы данных в режиме деперсонализации.

        :rtype: bool
        """
        return all(
            manager.is_in_d15n_mode()
            for manager in itervalues(self._managers)
        )

    def switch_on(self):
        """Включает режим деперсонализации во всех менеджерах."""
        for manager in itervalues(self._managers):
            manager.switch_on()

    def switch_off(self):
        """Отключает режим деперсонализации во всех менеджерах."""
        for manager in itervalues(self._managers):
            manager.switch_off()

    @property
    def d15n_mode(self):
        """Возвращает менеджеры контекста для режима деперсонализации.

        Для каждого из зарегистрированных менеджеров деперсонализации создает
        менеджер контекста :class:`~m3_d15n.utils.D15nMode` и объединяет их.
        """
        if PY2:
            # pylint: disable=dict-iter-method
            result = contextlib.nested(*(
                manager.d15n_mode
                for manager in self._managers.itervalues()
            ))
        else:
            stack = contextlib.ExitStack()
            for manager in self._managers.values():
                stack.enter_context(manager.d15n_mode)
            result = stack

        return result

    def add_field(self, field, function_name):
        """Добавляет поле в параметры деперсонализации.

        Менеджер деперсонализации выбирается с помощью роутера Django.

        :param field: поле модели.
        :type field: django.db.models.fields.Field

        :param basestring function_name: имя деперсонализирующей функции в БД,
            например ``'d15_core.depersonalize_name'``.
        """
        alias = router.db_for_write(field.model)
        manager = self._managers[alias]
        manager.add_field(field, function_name)

    def delete_field(self, field):
        """Удаляет поле модели из параметров деперсонализации.

        Менеджер деперсонализации выбирается с помощью роутера Django.

        :param field: поле модели
        :type field: django.db.models.fields.Field
        """
        alias = router.db_for_write(field.model)
        manager = self._managers[alias]
        manager.delete_field(field)

    def sync_config(self):
        """Выполняет синхронизацию параметров деперсонализации полей."""
        for manager in itervalues(self._managers):
            manager.sync_config()


class D15nViewsCleaner(object):

    """Удаляет деперсонализирующие представления на время применения миграций.

    Т.к. деперсонализирующие представления блокируют некоторые изменения в БД,
    перед выполнением management-команды ``migrate`` эти представления
    нужно удалить, а по завершении команды --- восстановить.
    """

    def __init__(self):  # noqa: D107
        from . import d15n_managers_registry
        self._app_names = {
            alias: set()
            for alias in connections
            if alias in d15n_managers_registry
        }

        pre_migrate.connect(self._pre_migrate_handler)
        post_migrate.connect(self._post_migrate_handler)

    def _pre_migrate_handler(self, app_config, using, **kwargs):
        if using not in self._app_names:
            return

        app_names = self._app_names[using]
        if not app_names and self._is_initialized(using):
            self._drop_d15n_views(using)

        app_names.add(app_config.name)

    def _post_migrate_handler(self, app_config, using, **kwargs):
        if using not in self._app_names:
            return

        app_names = self._app_names[using]
        if app_config.name in self._app_names:
            app_names.remove(app_config.name)

        if not app_names and not self._is_initialized(using):
            self._create_d15n_views(using)

    @staticmethod
    def _is_initialized(database_alias):
        """Возвращает True, если представление для FieldConfig есть в БД.

        :rtype: bool
        """
        with connections[database_alias].cursor() as cursor:
            cursor.execute("""
                select
                    1
                from
                    pg_class tables
                    inner join pg_catalog.pg_namespace schemas on (
                        tables.relnamespace = schemas.oid
                    )
                where
                    tables.relkind = 'v' and
                    schemas.nspname = '{core_schema}' and
                    tables.relname = 'columns_config_as_names'
            """.format(
                core_schema=CORE_SCHEMA,
            ))
            result = cursor.fetchone() is not None

        return result

    @staticmethod
    def _drop_d15n_views(database_alias):
        """Удаляет деперсонализирующие представления."""
        FieldConfig = apps.get_model('m3_d15n', 'FieldConfig')

        for table_name in FieldConfig.objects.using(
            database_alias
        ).values_list('table_name', flat=True).distinct():
            with connections[database_alias].cursor() as cursor:
                cursor.execute("""
                    drop view {data_schema}.{table_name}
                """.format(
                    data_schema=DATA_SCHEMA,
                    table_name=table_name,
                ))

        # Установка параметра in_migration, сообщающего триггеру on_sql_drop о
        # том, что выполняется процесс миграции.
        with connections[database_alias].cursor() as cursor:
            cursor.callproc('set_config', (
                CORE_SCHEMA + '.in_migration', 'TRUE', False
            ))

    @staticmethod
    def _create_d15n_views(database_alias):
        """Создает деперсонализирующие представления."""
        with connections[database_alias].cursor() as cursor:
            cursor.callproc('set_config', (
                CORE_SCHEMA + '.in_migration', '', False
            ))

            cursor.callproc(CORE_SCHEMA + '.update_views_set')
